// pepper_posix_native_tcp.cc - Native Pepper TCP implementation.

// Copyright 2013, 2014, 2015 Richard Woodbury
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "mosh_nacl/pepper_posix_native_tcp.h"

#include <errno.h>
#include <netinet/in.h>
#include <stdio.h>
#include <string.h>
#include <sys/uio.h>

#include "ppapi/c/pp_errors.h"
#include "ppapi/cpp/completion_callback.h"
#include "ppapi/cpp/module.h"
#include "ppapi/cpp/var.h"

namespace PepperPOSIX {

NativeTCP::NativeTCP(const pp::InstanceHandle& instance_handle)
    : socket_(new pp::TCPSocket(instance_handle)), factory_(this) {}

NativeTCP::~NativeTCP() {}

int NativeTCP::Bind(const pp::NetAddress& address) {
  pp::Var string_address = address.DescribeAsString(true);
  if (string_address.is_undefined()) {
    Log("NativeTCP::Bind(): Address is bogus.");
    return EFAULT;
  }

  return socket_->Bind(address, pp::CompletionCallback());
}

int NativeTCP::Connect(const pp::NetAddress& address) {
  address_ = address;
  pp::Var string_address = address_.DescribeAsString(true);
  if (string_address.is_undefined()) {
    Log("NativeTCP::Connect(): Address is bogus.");
    return EFAULT;
  }
  if (IsBlocking() == true) {
    Log("NativeTCP::Connect(): Not in non-blocking mode, not implemented!");
  }
  // API calls need to be done on the main thread.
  pp::Module::Get()->core()->CallOnMainThread(
      0, factory_.NewCallback(&NativeTCP::ConnectOnMainThread));
  return EINPROGRESS;
}

// This callback should only be called on the main thread.
void NativeTCP::ConnectOnMainThread(__attribute__((unused)) int32_t unused) {
  int32_t result =
      socket_->Connect(address_, factory_.NewCallback(&NativeTCP::Connected));
  if (result != PP_OK_COMPLETIONPENDING) {
    Log("NativeTCP::ConnectOnMainThread(): "
        "socket_->Connect() returned %d",
        result);
    // TODO(rpwoodbu): Perhaps crash here?
  }
  // TODO(rpwoodbu): Flesh out error mapping.
}

void NativeTCP::Connected(int32_t result) {
  if (result == PP_OK) {
    target_->UpdateWrite(true);
    StartReceive();
    return;
  }

  Log("NativeTCP::Connected(): Connection failed; result: %d", result);
  switch (result) {
    case PP_ERROR_NOACCESS:
      connection_errno_ = EACCES;
      break;
    case PP_ERROR_ADDRESS_UNREACHABLE:
      connection_errno_ = EHOSTUNREACH;
      break;
    case PP_ERROR_CONNECTION_REFUSED:
      connection_errno_ = ECONNREFUSED;
      break;
    case PP_ERROR_CONNECTION_TIMEDOUT:
      connection_errno_ = ETIMEDOUT;
      break;
    default:
      connection_errno_ = EIO;
      break;
  }

  target_->UpdateWrite(true);
  target_->UpdateRead(true);
}

ssize_t NativeTCP::Send(const void* buf, size_t count, int flags) {
  if (flags != 0) {
    Log("NativeTCP::Send(): Unsupported flag: 0x%x", flags);
  }
  int32_t result =
      socket_->Write((const char*)buf, count, pp::CompletionCallback());
  if (result < 0) {
    Log("NativeTCP::Send(): Got negative result: %d", result);
  }
  return result;
}

// StartReceive prepares to receive more data, and returns without blocking.
void NativeTCP::StartReceive() {
  int32_t result = socket_->Read(receive_buffer_, sizeof(receive_buffer_),
                                 factory_.NewCallback(&NativeTCP::Received));
  if (result != PP_OK_COMPLETIONPENDING) {
    Log("NativeTCP::StartReceive(): Read unexpectedly returned %d", result);
    // TODO(rpwoodbu): Perhaps crash here?
  }
}

// Received is the callback result of StartReceive().
void NativeTCP::Received(int32_t result) {
  if (result < 0) {
    Log("NativeTCP::Received(%d, ...): Negative result; bailing.", result);
    return;
  }
  AddData(receive_buffer_, result);
  // Await another packet.
  StartReceive();
}

// Close the socket.
int NativeTCP::Close() {
  // Destroying socket_ is the same as closing it.
  socket_.reset();
  return 0;
}

}  // namespace PepperPOSIX
